/* wydawca - automatic release submission daemon
   Copyright (C) 2007, 2008, 2009 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "save-cwd.h"
#if defined USE_SENDFILE
# if defined __linux__ && defined HAVE_SYS_SENDFILE_H
#  include <sys/sendfile.h>
# else
#  undef USE_SENDFILE
# endif
#endif

/* Return true if ARG is NULL or is a sub-directory of DIR */
int
sub_dir_p (char *arg, char *dir)
{
  int dlen;

  if (!arg)
    return 0;
    
  dlen = strlen (dir);

  return strlen (arg) > dlen
         && memcmp (dir, arg, dlen) == 0
         && arg[dlen] == '/';
}

/* Concatenate BASE directory name, a single directory separator (/) and
   NAME. Return in PBASELEN the length of BASE, not counting eventual trailing
   slash. */
char *
concat_dir (const char *base, const char *name, size_t *pbaselen)
{
  size_t len = strlen (base);
  size_t size;
  char *dir;
    
  while (len > 0 && base[len-1] == '/')
    len--;

  size = len + 1 + strlen (name);
  dir = xmalloc (size + 1);
  memcpy (dir, base, len);
  dir[len++] = '/';
  strcpy (dir + len, name);

  if (pbaselen)
    *pbaselen = len;
  return dir;
}

/* Create the directory DIR, eventually creating all intermediate directories
   starting from DIR + BASELEN, with owner UID and GID. */
int
create_hierarchy (char *dir, size_t baselen, uid_t uid, gid_t gid)
{
  int rc;
  struct stat st;
  char *p;
  
  if (stat (dir, &st) == 0)
    {
      if (!S_ISDIR (st.st_mode))
	{
	  logmsg (LOG_ERR, "component %s is not a directory", dir);
	  return 1;
	}
      return 0;
    }
  else if (errno != ENOENT)
    {
      logmsg (LOG_ERR, "cannot stat file %s: %s", dir, strerror (errno));
      return 1;
    }
  
  p = strrchr (dir, '/');
  if (p)
    {
      if (p - dir + 1 < baselen)
	{
	  logmsg (LOG_ERR, "base directory %s does not exist", dir);
	  return 1;
	}
      *p = 0;
    }

  rc = create_hierarchy (dir, baselen, uid, gid);
  if (rc == 0)
    {
      if (p)
	*p = '/';
      if (mkdir (dir, MKDIR_PERMISSIONS))
	{
	  logmsg (LOG_ERR, "cannot create directory %s: %s",
		  dir, strerror (errno));
	  rc = 1;
	}
      if (chown (dir, uid, gid))
	{
	  logmsg (LOG_NOTICE, "cannot change ownership of %s: %s",
		  dir, strerror (errno));
	}
    }
  return rc;
}

/* Create a directory BASE/NAME (with eventual intermediate directories in
   NAME). Use UID and GID as owner ids.
   Do nothing if dry_run_mode is set. */
char *
create_directory (const char *base, const char *name, uid_t uid, gid_t gid)
{
  size_t baselen;
  char *dir = concat_dir (base, name, &baselen);

  if (!dry_run_mode && create_hierarchy (dir, baselen, uid, gid))
    {
      free (dir);
      dir = NULL;
    }
  return dir;
}


/* Copy FILE to DST_FILE, creating the latter with owner UID and GID.
   Prefer sendfile(2), if available. Otherwise use plain old approach. */
int
copy_file (const char *file, const char *dst_file, uid_t uid, gid_t gid)
{
  int in_fd, out_fd;
  struct stat st;
  enum { copy_ok, copy_fallback, copy_failed } rc = copy_fallback;

  in_fd = open (file, O_RDONLY);

  if (in_fd == -1)
    {
      logmsg (LOG_ERR, "cannot open source file %s for reading: %s",
	      file, strerror (errno));
      return 1;
    }

  if (fstat (in_fd, &st))
    {
      logmsg (LOG_ERR, "cannot stat source file %s: %s",
	      file, strerror (errno));
      close (in_fd);
      return 1;
    }
  
  out_fd = creat (dst_file, CREAT_PERMISSIONS);
  if (out_fd == -1)
    {
      logmsg (LOG_ERR, "cannot create destination file %s: %s",
	      file, strerror (errno));
      close (in_fd);
      return 1;
    }
  
#ifdef USE_SENDFILE
  if (enable_sendfile)
    {
      off_t offset = 0;
      ssize_t count;
   
      count = sendfile (out_fd, in_fd, &offset, st.st_size);
      if (count == -1)
	{
	  if (errno == EINVAL || errno == ENOSYS)
	    rc = copy_fallback;
	  else
	    rc = copy_failed;
	  logmsg ((rc == copy_fallback) ? LOG_WARNING : LOG_ERR,
		  "sendfile: copying %s to %s failed: %s",
		  file, dst_file, strerror (errno));
	}
      else if (count < st.st_size)
	{
	  rc = copy_failed;
	  logmsg (LOG_ERR, "copying %s to %s failed: short write",
		  file, dst_file);
	}
      else
	rc = copy_ok;
    }

  if (rc == copy_fallback)
#endif
  {
    char *buf = NULL;
    size_t bufsize;
    size_t fsize = st.st_size;
    
    for (bufsize = fsize; bufsize > 0 && (buf = malloc (bufsize)) == NULL;
	 bufsize /= 2)
      ;
    if (bufsize == 0)
      xalloc_die ();

    rc = copy_ok;
    while (fsize > 0)
      {
	size_t rest;
	size_t rdbytes;
	
	rest = fsize > bufsize ? bufsize : fsize;
	rdbytes = read (in_fd, buf, rest);
	if (rdbytes == -1)
	  {
	    logmsg (LOG_ERR, "unexpected error reading %s: %s",
		    file, strerror (errno));
	    rc = copy_failed;
	    break;
	  }
	rest = write (out_fd, buf, rdbytes);
	if (rest == -1)
	  {
	    logmsg (LOG_ERR, "unexpected error writing to %s: %s",
		    dst_file, strerror (errno));
	    rc = copy_failed;
	    break;
	  }	   
	else if (rest != rdbytes)
	  {
	    logmsg (LOG_ERR, "short write on %s", dst_file);
	    rc = copy_failed;
	  }
	fsize -= rdbytes;
      }
    free (buf);
  }

  close (in_fd);
  close (out_fd);
  if (rc != copy_ok)
    {
      unlink (dst_file);
      return 1;
    }
  return 0;
}

/* Move FILE to DST_FILE. If they reside on different devices, use copy_file
   + unlink.
   UID and GID give DST_FILE ownership. */
int
do_move_file (const char *file, const char *dst_file, uid_t uid, gid_t gid)
{
  int rc = 0;

  if (rename (file, dst_file))
    {
      if (errno == EXDEV)
	{
	  if (copy_file (file, dst_file, uid, gid))
	    {
	      logmsg (LOG_CRIT, "cannot copy %s to %s: %s",
		      file, dst_file, strerror (errno));
	      rc = 1;
	    }
	  else if (unlink (file))
	    {
	      logmsg (LOG_ERR, "cannot unlink %s: %s",
		      file, strerror (errno));
	    }
	}
      else
	{
	  logmsg (LOG_CRIT, "cannot move %s to %s: %s",
		  file, dst_file, strerror (errno));
	  rc = 1;
	}
    }
  return rc;
}

/* Append the FILE to the tar ARCHIVE.
   Do nothing if dry_run_mode is set. */
int
tar_append_file (const char *archive, const char *file)
{
  const char *argv[6];
  
  if (debug_level)
    logmsg (LOG_DEBUG, "tarring %s to %s", file, archive);
  if (dry_run_mode)
    {
      UPDATE_STATS (STAT_ARCHIVES);
      return 0;
    }

  argv[0] = tar_command_name;
  argv[1] = "-f";
  argv[2] = archive;
  argv[3] = "-r";
  argv[4] = file;
  argv[5] = NULL;

  switch (wydawca_exec (6, argv, NULL))
    {
    case exec_success:
      UPDATE_STATS (STAT_ARCHIVES);
      return 0;

    case exec_fail:
    case exec_error:
      logmsg (LOG_ERR, "cannot archive %s", file);
      break;
    }

  return 1;
}

/* Backup a file to a directory.
   Arguments:
     DST_FILE        - File nam to back up.
     DST_DIR         - Directory part of DST_FILE.
     FILE            - File part of DST_FILE; can contain subdirs.
     ARCHIVE         - Archive descriptor.
     UID, GID        - Ownership
     RELDIR          - Directory part of FILE

   Do nothing if dry_run_mode is set.    */
int
backup_file (const char *dst_file, const char *dst_dir, const char *file,
	     struct archive_descr *archive, uid_t uid, gid_t gid,
	     const char *reldir)
{
  int rc = 0;
  char *adir;
  char *file_name;
  
  if (archive->name[0] == '/')
    adir = create_directory (archive->name, reldir, uid, gid);
  else
    adir = create_directory (dst_dir, archive->name, uid, gid);
  if (!adir)
    return 1;

  file_name = concat_dir (adir, file, NULL);
  if (access (file_name, F_OK) == 0)
    {
      if (archive->backup_type == no_backups)
	{
	  if (debug_level)
	    logmsg (LOG_DEBUG, "removing previous archive file `%s'",
		    file_name);
	  if (!dry_run_mode && unlink (file_name))
	    {
	      logmsg (LOG_ERR, "cannot unlink previous archive file `%s': %s",
		      file_name, strerror (errno));
	      free (file_name);
	      free (adir);
	      return 1;
	    }
	}
      else
	{
	  char *archive_file_name =
	    find_backup_file_name (file_name, archive->backup_type);
	  if (debug_level)
	    logmsg (LOG_DEBUG, "backing up previous archive file `%s' to `%s'",
		    file_name, archive_file_name);
	  if (!dry_run_mode)
	    {
	      rc = do_move_file (file_name, archive_file_name, uid, gid);
	      if (rc)
		{
		  logmsg (LOG_ERR, "backing `%s' up as `%s' failed: %s",
			  file_name, archive_file_name, strerror (errno));
		  free (archive_file_name);
		  free (file_name);
		  free (adir);
		  return 1;
		}
	    }
	  free (archive_file_name);
	}
    }

  if (debug_level)
    logmsg (LOG_DEBUG, "archiving `%s' to `%s'",  dst_file, file_name);
  if (!dry_run_mode)
    {
      rc = do_move_file (dst_file, file_name, uid, gid);
      if (rc)
	logmsg (LOG_ERR, "archiving `%s' as `%s' failed: %s",
		dst_file, file_name, strerror (errno));
    }
  free (file_name);
  free (adir);
  if (rc == 0)
    UPDATE_STATS (STAT_ARCHIVES);
  return rc;
}	      

/* Select the appropriate backup type and backup a file. See backup_file
   for the argument description. */
int
do_archive_file (const char *dst_file, const char *dst_dir, const char *file,
		 struct archive_descr *archive, uid_t uid, gid_t gid,
		 const char *reldir)
{
  switch (archive->type)
    {
    case archive_none:
      break;

    case archive_directory:
      return backup_file (dst_file, dst_dir, file, archive, uid, gid, reldir);
      
    case archive_tar:
      return tar_append_file (archive->name, dst_file);
    }
  return 0;
}

/* Move the part FILE_ID of the triplet TRP between the directories in
   DPAIR. RELDIR gives relative directory (i.e. the directory part of
   the file name) for backup purposes.

   Do nothing if dry_run_mode is set. */
int
move_file (struct file_triplet *trp, struct directory_pair *dpair,
	   enum file_type file_id, const char *reldir)	  
{
  char *dst_file;
  int rc = 0;
  char *dst_dir = create_directory (dpair->dest_dir, reldir,
				    TRIPLET_UID (trp), TRIPLET_GID (trp));
  
  if (!dst_dir)
    return 1;
  dst_file = concat_dir (dst_dir, trp->file[file_id].name, NULL);
  
  if (debug_level)
    logmsg (LOG_DEBUG, "installing %s to %s", trp->file[file_id].name,
	    dst_dir);

  if (access (dst_file, F_OK) == 0)
    rc = do_archive_file (dst_file, dst_dir, trp->file[file_id].name,
			  &dpair->archive,
			  TRIPLET_UID (trp), TRIPLET_GID (trp), reldir);
  
  if (!dry_run_mode && rc == 0)
    rc = do_move_file (trp->file[file_id].name, dst_file,
		       TRIPLET_UID (trp), TRIPLET_GID (trp));
      
  free (dst_file);
  free (dst_dir);
  if (rc == 0)
    UPDATE_STATS (STAT_UPLOADS);
  return rc;
}

/* Archive the part FILE_ID of the triplet TRP located in DPAIR->dest_dir.
   Do nothing if dry_run_mode is set.

   FIXME: Remove the file afterwards?
*/
int
archive_file (struct file_triplet *trp, struct directory_pair *dpair,
	      const char *file_name, const char *reldir)	  
{
  char *dst_file;
  int rc = 0;
  char *dst_dir = create_directory (dpair->dest_dir, reldir,
				    TRIPLET_UID (trp), TRIPLET_GID (trp));
  
  if (!dst_dir)
    return 1;

  dst_file = safe_file_name (concat_dir (dst_dir, file_name, NULL));
  if (!sub_dir_p (dst_file, dpair->dest_dir))
    {
      logmsg (LOG_ERR, "file to be archived `%s' does not lie under `%s'",
	      dst_file, dpair->dest_dir);
      free (dst_file);
      free (dst_dir);
      return 1;
    }
  
  if (access (dst_file, F_OK) == 0)
    {
      if (debug_level)
	logmsg (LOG_DEBUG, "Archiving file `%s'", dst_file);
      rc = do_archive_file (dst_file, dst_dir, file_name, &dpair->archive,
			    TRIPLET_UID (trp), TRIPLET_GID (trp), reldir);
    }
  else
    logmsg (LOG_NOTICE, "Nothing to archive: file `%s' does not exist",
	    dst_file);
  
  free (dst_file);
  free (dst_dir);
  if (rc == 0)
    UPDATE_STATS (STAT_ARCHIVES);
  return rc;
}

/* Create a symbolic link from WANTED_SRC to WANTED_DST in the subdirectory
   RELDIR of DPAIR->dest_dir. Get ownership information from TRP.

   Do nothing if dry_run_mode is set. */
int
symlink_file (struct file_triplet *trp, struct directory_pair *dpair,
	      const char *reldir,
	      const char *wanted_src, const char *wanted_dst)
{
  int rc = 0;
  struct saved_cwd cwd;
  char *dst_dir = create_directory (dpair->dest_dir, reldir,
				    TRIPLET_UID (trp), TRIPLET_GID (trp));
  char *src, *dst;
  
  if (!dst_dir)
    return 1;

  if (save_cwd (&cwd))
    {
      logmsg (LOG_ERR, "cannot save current directory: %s",
	      strerror (errno));
      return 1;
    }
  
  src = safe_file_name_alloc (wanted_src);
  if (!src || src[0] == '/')
    {
      logmsg (LOG_ERR, "symlink source `%s' does not lie under `%s'",
	      wanted_src, dpair->dest_dir);
      free (src);
      return 1;
    }

  dst = safe_file_name_alloc (wanted_dst);
  if (!dst || dst[0] == '/')
    {
      logmsg (LOG_ERR, "symlink destination `%s' does not lie under `%s'",
	      wanted_dst, dpair->dest_dir);
      free (src);
      free (dst);
      return 1;
    }
  
  if (debug_level)
    logmsg (LOG_DEBUG, "symlinking %s to %s in directory %s",
	    src, dst, dst_dir);
  
  if (!dry_run_mode)
    {
      char *p = strrchr (dst, '/');
      if (p > dst)
	{
	  char *dir;
	  
	  *p = 0;
	  dir = create_directory (dpair->dest_dir, dst, 
				  TRIPLET_UID (trp), TRIPLET_GID (trp));
	  if (!dir)
	    rc = 1;
	  else
	    free (dir);
	  *p = '/';
	}
      
      if (rc == 0)
	{
	  if (chdir (dst_dir))
	    logmsg (LOG_ERR, "cannot change to %s: %s",
		    dst_dir, strerror (errno));
	  else
	    {
	      struct stat st;
	      
	      if (lstat (dst, &st) == 0)
		{
		  if (!S_ISLNK (st.st_mode))
		    {
		      logmsg (LOG_ERR,
			      "file %s exists and is not a symbolic link",
			      dst, strerror (errno));
		      rc = 1;
		    }
		  else if (unlink (dst))
		    {
		      logmsg (LOG_ERR,
			      "Cannot unlink %s: %s",
			      dst, strerror (errno));
		      rc = 1;
		    }
		}
	      else if (errno != ENOENT)
		{
		  logmsg (LOG_ERR,
			  "cannot stat file %s: %s",
			  dst, strerror (errno));
		  rc = 1;
		}

	      if (rc == 0)
		{
		  rc = symlink (src, dst);
		  if (rc)
		    logmsg (LOG_ERR,
			    "symlinking %s to %s in directory %s failed: %s",
			    src, dst, dst_dir, strerror (errno));
		}
	    }
	}
    }

  if (restore_cwd (&cwd))
    {
      logmsg (LOG_EMERG, "cannot restore current directory: %s",
	      strerror (errno));
      exit (1);
    }

  free (src);
  free (dst);
  free (dst_dir);
  if (rc == 0)
    UPDATE_STATS (STAT_SYMLINKS);
  return rc;
}

/* Auxiliary function for rmsymlink_file (see below) */
static int
do_rmsymlink_file (const char *dst_file)
{
  struct stat st;
  
  if (debug_level)
    logmsg (LOG_DEBUG, "Removing symbolic link %s", dst_file);
  
  if (stat (dst_file, &st))
    {
      if (errno == ENOENT)
	{
	  logmsg (LOG_NOTICE, "Symlink `%s' does not exist", dst_file);
	  return 0;
	}
      if (!S_ISLNK (st.st_mode))
	{
	  logmsg (LOG_ERR, "Refusing to unlink %s: is not a symlink",
		  dst_file);
	  return 1;
	}
    }
  if (dry_run_mode)
    return 0;
  if (unlink (dst_file))
    {
      logmsg (LOG_ERR, "Cannot unlink %s: %s", dst_file, strerror (errno));
      return 1;
    }
  return 0;
}

/* Remove the symbolic link DPAIR->dest_dir/RELDIR/FILE_NAME
   
   Get ownership information from TRP.

   Do nothing if dry_run_mode is set. */
int
rmsymlink_file (struct file_triplet *trp, struct directory_pair *dpair,
		const char *reldir, const char *file_name)
{
  char *dst_file;
  int rc = 0;
  char *dst_dir = create_directory (dpair->dest_dir, reldir,
				    TRIPLET_UID (trp), TRIPLET_GID (trp));

  if (!dst_dir)
    return 1;

  dst_file = safe_file_name (concat_dir (dst_dir, file_name, NULL));
  if (!sub_dir_p (dst_file, dpair->dest_dir))
    {
      logmsg (LOG_ERR, "refusing to remove a symlink `%s' that is not "
	      "located under `%s'",
	      dst_file, dpair->dest_dir);
      free (dst_file);
      free (dst_dir);
      return 1;
    }

  rc = do_rmsymlink_file (dst_file);

  free (dst_file);
  free (dst_dir);
  if (rc == 0)
    UPDATE_STATS (STAT_RMSYMLINKS);
  return rc;
}
  
